package jigcell.project;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * Backing store for a data file on disk.
 *
 * This code is licensed under the DARPA BioCOMP Open Source License.  See LICENSE for more details.
 *
 * @author Nicholas Allen
 */

class DiskStore implements ProjectFileStore {

   /**
    * Storage information key for the project directory
    */

   public final static String KEY_DISKSTORE_PROJECTDIRECTORY = "diskstore_projectdirectory";

   /**
    * Real storage for this data file
    */

   protected File file;

   /**
    * Project containing this data file
    */

   protected Project project;

   /**
    * Name of this data file
    */

   protected String name;

   /**
    * Version of this data file
    */

   protected String version;

   /**
    * Creates the store data needed to use disk storage.
    *
    * @param projectDirectory Base directory in which project data files should be located
    *
    * @throws ProjectException if an IO error occurs
    */

   public final static Map createDiskStoreData (File projectDirectory) throws ProjectException {
      Map map = new HashMap ();
      try {
         map.put (KEY_DISKSTORE_PROJECTDIRECTORY, projectDirectory.getCanonicalPath ());
      } catch (Exception e) {
         throw new ProjectException (e);
      }
      return map;
   }

   /**
    * Opens the storage for a data file.  If no data file exists, a new one should be created.  The name and version must support at least [a-zA-Z0-9_-' ].
    *
    * @param name Name of the data file
    * @param version Version of the data file
    * @param project Project
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   public DiskStore (String name, String version, Project project) throws ProjectException {
      this.name = name;
      this.version = version;
      this.project = project;
      file = createFile (name, version, project);
   }

   /**
    * Deletes the storage for this data file.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized void delete () throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      name = null;
      version = null;
      project = null;
      File oldFile = file;
      file = null;
      if (!oldFile.delete ()) throw new ProjectException ("Error during deletion");
   }

   /**
    * The time at which this data file was last modified.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized Date getModificationDate () throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      try {
         return new Date (file.lastModified ());
      } catch (Exception e) {
         throw new ProjectException (e);
      }
   }

   /**
    * Stream from which this data file can be read.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized InputStream getInputStream () throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      try {
         return new BufferedInputStream (new FileInputStream (file));
      } catch (Exception e) {
         throw new ProjectException (e);
      }
   }

   /**
    * Stream to which this data file can be written.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized OutputStream getOutputStream () throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      try {
         return new BufferedOutputStream (new FileOutputStream (file));
      } catch (Exception e) {
         throw new ProjectException (e);
      }
   }

   /**
    * Changes the name of this data file.  The name must support at least [a-zA-Z0-9_-' ].
    *
    * @param name New name for the data file
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   public synchronized void setName (String name) throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      if (!Project.isValidName (name)) throw new ProjectException ("Invalid name " + name);
      File newFile = createFile (name, version, project);
      if (!file.renameTo (newFile)) throw new ProjectException ("Unable to rename");
      file = newFile;
   }


   /**
    * Changes the project of this data file.
    *
    * @param project New project for the data file
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized void setProject (Project project) throws ProjectException {
      if (file == null) throw new ProjectException ("Storage has been destroyed");
      File newFile = createFile (name, version, project);
      if (!file.renameTo (newFile)) throw new ProjectException ("Unable to rename");
      file = newFile;
   }

   /**
    * Builds a disk file for this data file.
    *
    * @param name Name of the data file
    * @param version Version of the data file
    * @param project Project containing the data file
    *
    * @throws ProjectException if an IO error occurs
    */

   protected File createFile (String name, String version, Project project) throws ProjectException {
      String path = (String) project.getStoreData ().get (KEY_DISKSTORE_PROJECTDIRECTORY);
      if (path == null) { System.out.println("Not a valid path"); throw new ProjectException ("Project not configured for disk storage");}
      try {
System.out.println("Trying to create");
         File tempf = new File (path + File.separatorChar + name + "!" + version);
System.out.println("Name set at: " + tempf.getCanonicalPath());
         tempf.createNewFile();
         return tempf.getCanonicalFile ();
      } catch (Exception e) {
         throw new ProjectException (e);
      }
   }
}
