package jigcell.project;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A Project is a collection of data files that can be read, modified, and are backed by some permanent storage mechanism.  Projects should never be cached or otherwise stored for extended
 * periods of time by applications.
 *
 * This code is licensed under the DARPA BioCOMP Open Source License.  See LICENSE for more details.
 *
 * @author Nicholas Allen
 */

public class Project {

   /**
    * Next ID number to assign
    */

   protected static int nextID = 0;

   /**
    * Pattern for valid file names
    */

   protected final static Matcher matcher = Pattern.compile ("[\\w-' ]+").matcher (null);

   /**
    * Constructor to instantiate an instance of the data file storage for this project
    */

   protected Constructor storeConstructor;

   /**
    * ID number of this project
    */

   protected int id;

   /**
    * Storage information
    */

   protected Map storeData;

   /**
    * Storage for index data
    */

   protected ProjectFileStore index;

   /**
    * Name of this project
    */

   protected String name;

   /**
    * Information about this project.
    */

   protected static class ProjectInfo implements Serializable {

      /**
       * Description of the project
       */

      public String description;

      /**
       * Names of the data files in this project
       */

      public List names;

      /**
       * Creates a new record for a project.
       */

      public ProjectInfo () {
         names = new ArrayList ();
      }

      /**
       * All of the data files in this project.  Every element of the returned list will be of type ProjectFile.
       *
       * @param project Project
       */

      public synchronized List getProjectFiles (Project project) throws ProjectException {
         List projectFiles = new ArrayList ();
         String name, projectName = project.getName ();
         for (Iterator iterator = names.iterator (); iterator.hasNext (); ) if (!projectName.equals (name = (String) iterator.next ())) projectFiles.add (new ProjectFile (name, project, false));
         return Collections.unmodifiableList (projectFiles);
      }
   }

   static boolean isValidName (String name) {
      if (name == null || !name.equals (name.trim ())) return false;
      boolean result = false;
      synchronized (matcher) {
         result = matcher.reset (name).matches ();
         matcher.reset ();
      }
      return result;
   }

   /**
    * Permanently deletes this project.  Further operations on a deleted project will fail.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized void delete () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      ProjectException failed = null;
      try {
         for (Iterator iterator = readIndex ().getProjectFiles (this).iterator (); iterator.hasNext (); ) try {
            ((ProjectFile) iterator.next ()).delete ();
         } catch (ProjectException e) {
            failed = e;
         }
         index.delete ();
      } catch (ProjectException e) {
         failed = e;
      }
      storeConstructor = null;
      storeData = null;
      index = null;
      name = null;
      if (failed != null) throw failed;
   }

   /**
    * The text describing this project.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized String getDescription () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return readIndex ().description;
   }

   /**
    * The time at which this project was last modified.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized Date getModificationDate () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return index.getModificationDate ();
   }

   /**
    * The name of this project.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized String getName () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return name;
   }

   /**
    * All of the data files in this project.  Every element of the returned list will be of type ProjectFile.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized List getProjectFiles () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return readIndex ().getProjectFiles (this);
   }

   /**
    * Sets the text describing this project.
    *
    * @param description New description for this project
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized void setDescription (String description) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      ProjectInfo info = readIndex ();
      info.description = description;
      writeIndex (info);
   }

   /**
    * Sets the name of this project.  The name must be unique, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and last characters must not be a space.
    *
    * @param name New name for this project
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   public synchronized void setName (String name) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (!isValidName (name)) throw new ProjectException ("Invalid name " + name);
   }

   /**
    * Reads the index information for this data file.
    *
    * @throws ProjectException if an IO error occurs
    */

   protected ProjectInfo readIndex () throws ProjectException {
      ObjectInputStream in = null;
      try {
         return (ProjectInfo) (in = new ObjectInputStream (index.getInputStream ())).readObject ();
      } catch (Exception e) {
         throw new ProjectException (e);
      } finally {
         if (in != null) try {
            in.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Writes the index information for this project.
    *
    * @param info Index information for this project
    *
    * @throws ProjectException if an IO error occurs
    */

   protected void writeIndex (ProjectInfo info) throws ProjectException {
      ObjectOutputStream out = null;
      try {
         (out = new ObjectOutputStream (index.getOutputStream ())).writeObject (info);
      } catch (Exception e) {
         throw new ProjectException (e);
      } finally {
         if (out != null) try {
            out.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Creates a new project.  The name of the project must be unique, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and last characters must not be a space.
    *
    * @param name Name of new project
    * @param storeClass Class to use for data file storage.  This class must be an instance of ProjectFileStore.
    * @param storeData Configuration data needed by the storage class.  The storage class should provide a method for creating this data.
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   Project (String name, Class storeClass, Map storeData) throws ProjectException {
      synchronized (Project.class) {
         id = nextID++;
      }
      this.name = name;
      try {
         if ((storeConstructor = storeClass.getDeclaredConstructor (new Class [] {String.class, String.class, Project.class})) == null)
            throw new ProjectException ("No usable constructor in storage class");
      } catch (Exception e) {
         throw new ProjectException (e);
      }
      this.storeData = Collections.unmodifiableMap (storeData);
      index = createStore (name, "index", false);
      try {
         readIndex();
      } catch (Exception e) {writeIndex (new ProjectInfo());}
   }

   /**
    * Renames a data file.  The name of the data file must be unique within this project, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and last characters must not be a
    * space.
    *
    * @param projectFile Data file
    * @param name New name for the data file
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   synchronized void changeProjectFileName (ProjectFile projectFile, String name) throws ProjectException {
      ProjectInfo info = readIndex ();
      if (info.names.contains (name)) throw new ProjectException ("A data file with that name already exists");
      if (!isValidName (name)) throw new ProjectException ("Invalid name " + name);
      if (projectFile.getProject () == this) info.names.remove (projectFile.getName ());
      info.names.add (name);
   }

   /**
    * Opens the storage for a data file.  If no data file exists, a new one should be created.  The name and version must support at least [a-zA-Z0-9_-' ].
    *
    * @param name Name of the data file
    * @param version Version of the data file
    * @param validate Whether to validate the uniqueness of the name
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   synchronized ProjectFileStore createStore (String name, String version, boolean validate) throws ProjectException {
      if (index != null) {
	ProjectInfo info = readIndex ();
      	if (info.names.contains (name)) {
         	if (validate) throw new ProjectException ("A data file with that name already exists");
      	} else {
           info.names.add (name);
           writeIndex(info);
        }
      }
System.out.println("Creating/loading store for data file");
      try {
         return (ProjectFileStore) storeConstructor.newInstance (new Object [] {name, version, this});
      } catch (Exception e) {
System.out.println("Exception caught while creating new instance of store");
         throw new ProjectException (e);
      }
   }

   /**
    * The ID number for this project.
    */

   int getProjectID () {
    return id;
   }

   /**
    * Storage information for this project.
    */

   Map getStoreData () {
      return storeData;
   }
}
