package jigcell.project;

import java.beans.ExceptionListener;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;

/**
 * A ProjectFile is a data file that can be read, modified, and is backed by some permanent storage mechanism.  Project files should never be cached or otherwise stored for extended periods
 * of time by applications.
 *
 * This code is licensed under the DARPA BioCOMP Open Source License.  See LICENSE for more details.
 *
 * @author Nicholas Allen
 */

public class ProjectFile {

   /**
    * Map from MIME types to handlers
    */

   protected final static Map handlers = Collections.synchronizedMap (new HashMap ());

   /**
    * Project containing this data file
    */

   protected Project project;

   /**
    * Storage for index data
    */

   protected ProjectFileStore index;

   /**
    * Name of this data file
    */

   protected String name;

   /**
    * Information about a version of the data file
    */

   protected static class ProjectFileInfo implements Serializable {

      /**
       * Description of the data file
       */

      public String description;

      /**
       * MIME type of the data file
       */

      public String mimeType;

      /**
       * Version of the data file
       */

      public String version;

      /**
       * Backing store for the data file
       */

      protected transient ProjectFileStore store;

      /**
       * Creates a new record for a version of the data file.
       *
       * @param mimeType Encoding for the data file
       * @param description Description for the data file
       * @param version Version of the data file
       */

      public ProjectFileInfo (String mimeType, String description, String version) {
         this.mimeType = mimeType;
         this.description = description;
         this.version = version;
      }

      /**
       * The backing store for a data file.
       *
       * @param projectFile Data file that this store is part of
       *
       * @throws ProjectException if an IO error occurs
       */

      public synchronized ProjectFileStore getStore (ProjectFile projectFile) throws ProjectException {
         if (store != null) return store;
         return store = projectFile.getProject ().createStore (projectFile.name, version, false);
      }
   }

   /**
    * Registers the handler for a data format.
    *
    * @param mimeType Data encoding the handler can process
    * @param handler Handler
    */

   public static void registerHandler (String mimeType, ProjectFileHandler handler) {
      if (mimeType == null || handler == null) throw new NullPointerException ();
      synchronized (handlers) {
         if (handlers.containsKey (mimeType)) throw new IllegalArgumentException ("Handler already defined for " + mimeType);
         handlers.put (mimeType, handler);
      }
   }

   /**
    * Creates a new data file and adds it to a project.  The name of the data file must be unique within this project, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and
    * last characters must not be a space.
    *
    * @param name Name of new data file
    * @param project Project
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   public ProjectFile (String name, Project project) throws ProjectException {
      this (name, project, true);
   }

   /**
    * Permanently deletes this data file.  Further operations on a deleted data file will fail.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized void delete () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      ProjectException failed = null;
      Iterator iterator = readIndex ().iterator ();
      for (ProjectFileInfo info; iterator.hasNext (); ) try {
         if ((info = (ProjectFileInfo) iterator.next ()) != null) info.getStore (this).delete ();
      } catch (ProjectException e) {
         failed = e;
      }
      try {
         index.delete ();
      } catch (ProjectException e) {
         failed = e;
      }
      project = null;
      index = null;
      name = null;
      if (failed != null) throw failed;
   }

   /**
    * The text describing a version of this project.
    *
    * @param version Version
    *
    * @throws ProjectException if an IO error occurs or the version is invalid
    */

   public synchronized String getDescription (int version) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (version < 1 || version > getVersion ()) throw new ProjectException ("Bad version");
      ProjectFileInfo info = (ProjectFileInfo) readIndex ().get (version - 1);
      if (info == null) throw new ProjectException ("No data for version " + version);
      return info.description;
   }

   /**
    * The data encoding for a version of this data file.
    *
    * @param version Version
    *
    * @throws ProjectException if an IO error occurs or the version is invalid
    */

   public synchronized String getMIMEType (int version) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (version < 1 || version > getVersion ()) throw new ProjectException ("Bad version");
      ProjectFileInfo info = (ProjectFileInfo) readIndex ().get (version - 1);
      if (info == null) throw new ProjectException ("No data for version " + version);
      return info.mimeType;
   }

   /**
    * The time at which a version of this data file was last modified.
    *
    * @param version Version
    *
    * @throws ProjectException if an IO error occurs or the version is invalid
    */

   public synchronized Date getModificationDate (int version) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (version < 1 || version > getVersion ()) throw new ProjectException ("Bad version");
      ProjectFileInfo info = (ProjectFileInfo) readIndex ().get (version - 1);
      if (info == null) throw new ProjectException ("No data for version " + version);
      return info.getStore (this).getModificationDate ();
   }

   /**
    * The name of this data file.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized String getName () throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return name;
   }

   public String toString() {
      try {return getName();} catch (Exception e) {e.printStackTrace(System.err); return new String();}
   }

   /**
    * The current version of this data file.  If this data file has never been saved, the current version is 0.
    *
    * @throws ProjectException if an IO error occurs
    */

   public synchronized int getVersion ()  throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      return readIndex ().size ();
   }

   /**
    * Retrieves a version of the data for this file.
    *
    * @param listener Listener to notify of any recoverable exceptions (may be null)
    * @param version Version
    *
    * @throws ProjectException if no data can be loaded
    */

   public synchronized Object load (ExceptionListener listener, int version) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (version < 1 || version > getVersion ()) throw new ProjectException ("Bad version");
      ProjectFileInfo info = (ProjectFileInfo) readIndex ().get (version - 1);
      if (info == null) throw new ProjectException ("No data for version " + version);
      String type = info.mimeType;
      ProjectFileHandler handler = (ProjectFileHandler) handlers.get (type);
      if (handler == null) throw new ProjectException ("No handler for data type " + type);
      InputStream in = null;
      try {
         return handler.load (in = info.getStore (this).getInputStream (), listener, type);
      } finally {
         if (in != null) try {
            in.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Moves this data file to another project.
    *
    * @param project Project to move to
    *
    * @throws ProjectException if an IO error occurs or if this data file cannot be moved
    */

   public synchronized void moveProjectFile (Project project) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (project == null || project == this.project) throw new ProjectException ("Bad project");
      Project innerProject, outerProject;
      if (project.getProjectID () < this.project.getProjectID ()) {
         innerProject = project;
         outerProject = this.project;
      } else {
         innerProject = this.project;
         outerProject = project;
      }
      synchronized (outerProject) {
         synchronized (innerProject) {
            project.changeProjectFileName (this, name);
            this.project = project;
            ProjectException failed = null;
            try {
               index.setProject (project);
            } catch (ProjectException e) {
               failed = e;
            }
            Iterator iterator = readIndex ().iterator ();
            for (ProjectFileInfo info; iterator.hasNext (); ) try {
               if ((info = (ProjectFileInfo) iterator.next ()) != null) info.getStore (this).setProject (project);
            } catch (ProjectException e) {
               failed = e;
            }
            if (failed != null) throw failed;
         }
      }
   }

   /**
    * Stores the data for this file.  After saving, the new data is the current version of this data file.
    *
    * @param listener Listener to notify of any recoverable exceptions (may be null)
    * @param description Description for this version of the data file
    * @param mimeType Data encoding for this data file
    * @param o Object to store
    *
    * @throws ProjectException if no data can be saved
    */

   public synchronized void save (ExceptionListener listener, String description, String mimeType, Object o) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      ProjectFileHandler handler = (ProjectFileHandler) handlers.get (mimeType);
      if (handler == null) throw new ProjectException ("No handler for data type " + mimeType);
      ProjectFileInfo info = new ProjectFileInfo (mimeType, description, String.valueOf (getVersion () + 1));
      OutputStream out = null;
      try {
         handler.save (out = info.getStore (this).getOutputStream (), listener, mimeType, o);
         List stores = readIndex ();
         stores.add (info);
         writeIndex (stores);
      } finally {
         if (out != null) try {
            out.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Renames this data file.  The name of the data file must be unique within its project, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and last characters must not be a
    * space.
    *
    * @param name New name for this data file
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   public synchronized void setName (String name) throws ProjectException {
      if (index == null) throw new ProjectException ("Storage has been destroyed");
      if (!Project.isValidName (name)) throw new ProjectException ("Invalid name " + name);
      synchronized (project) {
         project.changeProjectFileName (this, name);
         this.name = name;
         ProjectException failed = null;
         try {
            index.setName (name);
         } catch (ProjectException e) {
            failed = e;
         }
         Iterator iterator = readIndex ().iterator ();
         for (ProjectFileInfo info; iterator.hasNext (); ) try {
            if ((info = (ProjectFileInfo) iterator.next ()) != null) info.getStore (this).setName (name);
         } catch (ProjectException e) {
            failed = e;
         }
         if (failed != null) throw failed;
      }
   }

   /**
    * Reads the index information for this data file.  Returns the index information for each version of this data file as a list of ProjectFileInfo.
    *
    * @throws ProjectException if an IO error occurs
    */

   protected List readIndex () throws ProjectException {
      ObjectInputStream in = null;
      try {
         return (List) (in = new ObjectInputStream (index.getInputStream ())).readObject ();
      } catch (Exception e) {
         throw new ProjectException (e);
      } finally {
         if (in != null) try {
            in.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Writes the index information for this data file.
    *
    * @param stores Index information for each version of this data file
    *
    * @throws ProjectException if an IO error occurs
    */

   protected void writeIndex (List stores) throws ProjectException {
      ObjectOutputStream out = null;
      try {
         (out = new ObjectOutputStream (index.getOutputStream ())).writeObject (stores);
      } catch (Exception e) {
         throw new ProjectException (e);
      } finally {
         if (out != null) try {
            out.close ();
         } catch (Exception e) {}
      }
   }

   /**
    * Creates a new data file or opens an old data file.  The name of the data file must be unique within this project, non-empty, and is restricted to [a-zA-Z0-9_-' ].  The first and last
    * characters must not be a space.
    *
    * @param name Name of new data file
    * @param project Project
    * @param validate Whether to validate the uniqueness of the name
    *
    * @throws ProjectException if the name is invalid or an IO error occurs
    */

   ProjectFile (String name, Project project, boolean validate) throws ProjectException {
      if (validate && !Project.isValidName (name)) throw new ProjectException ("Invalid name " + name);
      this.name = name;
      this.project = project;
      index = project.createStore (name, "index", validate);
      try { readIndex (); } catch (Exception e) { writeIndex (new java.util.ArrayList());}

   }

   /**
    * Project this data file is contained in.
    */

   public Project getProject () {
      return project;
   }
}
