package jigcell.project;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;

/**
 * Backing store for a data file.
 *
 * All implementations of this interface should have an accessible constructor:
 *
 * Opens the storage for a data file.  If no data file exists, a new one should be created.  The name and version must support at least [a-zA-Z0-9_-' ].
 *
 * &amp;param name Name of the data file
 * &amp;param version Version of the data file
 * &amp;param project Project
 *
 * &amp;throws ProjectException if the name is invalid or an IO error occurs
 *
 * public (String name, String version, Project project) throws ProjectException;
 *
 * All implementations of this interface should have a factory method for creating the store data used by jigcell.project.Project.
 *
 * This code is licensed under the DARPA BioCOMP Open Source License.  See LICENSE for more details.
 *
 * @author Nicholas Allen
 */

interface ProjectFileStore {

   /**
    * Deletes the storage for this data file.
    *
    * @throws ProjectException if an IO error occurs
    */

   void delete () throws ProjectException;

   /**
    * The time at which this data file was last modified.
    *
    * @throws ProjectException if an IO error occurs
    */

   Date getModificationDate () throws ProjectException;

   /**
    * Stream from which this data file can be read.
    *
    * @throws ProjectException if an IO error occurs
    */

   InputStream getInputStream () throws ProjectException;

   /**
    * Stream to which this data file can be written.
    *
    * @throws ProjectException if an IO error occurs
    */

   OutputStream getOutputStream () throws ProjectException;

   /**
    * Changes the name of this data file.  The name must support at least [a-zA-Z0-9_-' ].
    *
    * @param name New name for the data file
    *
    * @throws ProjectException if an IO error occurs
    */

   void setName (String name) throws ProjectException;

   /**
    * Changes the project of this data file.
    *
    * @param project New project for the data file
    *
    * @throws ProjectException if an IO error occurs
    */

   void setProject (Project project) throws ProjectException;
}
